package net.stepniak.api.picheese.controller;

import net.stepniak.api.error.exception.ServerResponseException;
import net.stepniak.api.picheese.logic.SessionLogic;
import net.stepniak.api.picheese.logic.UserLogic;
import net.stepniak.picheese.user.AccountType;
import net.stepniak.api.picheese.repository.SessionDAO;
import net.stepniak.api.picheese.repository.UserDAO;
import net.stepniak.api.picheese.services.PicheeseBaseService;
import net.stepniak.api.picheese.validator.UserValidator;
import net.stepniak.picheese.error.server.ServerErrorType;
import org.codehaus.jettison.json.JSONException;
import org.codehaus.jettison.json.JSONObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.TransactionSystemException;

import javax.ws.rs.*;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.UriBuilder;
import java.security.NoSuchAlgorithmException;

@Path("auth")
//@Produces(MediaType.APPLICATION_JSON + "; charset=utf-8")
@Produces(MediaType.APPLICATION_JSON)
@Service
public class AuthController extends PicheeseBaseService<SessionLogic> {
  static Logger logger = LoggerFactory.getLogger(AuthController.class);

  public AuthController() {
    logger.info("AuthController constructor()");
  }

  public SessionDAO getSessionDAO() {
    return (SessionDAO) getDao();
  }

  @Autowired
  public void setSessionDAO(SessionDAO dao) {
    setDao(dao);
  }

  @Autowired
  private UserDAO userDAO;

  @POST
  @Consumes(MediaType.APPLICATION_FORM_URLENCODED)
  public Response createByForm(
    @FormParam("key") String key,
    @FormParam("email") String email,
    @FormParam("userName") String userName
  ) {
    logger.info("AuthController createByForm(\"{}\", \"{}\", \"{}\")", key, email, userName);
    return auth(key, email, userName);
  }

  @POST
  @Consumes(MediaType.APPLICATION_JSON)
  public Response createByJson(
    JSONObject jsonObject
  ) {

    String key = null;
    String email = null;
    String userName = null;

    try {
      key = jsonObject.getString("key");
      email = jsonObject.getString("email");
      userName = jsonObject.getString("userName");
    } catch (JSONException e) {
      //do nothing (email and userName are optional
    }

    logger.info("AuthController createByJson(\"{}\", \"{}\", \"{}\")", key, email, userName);
    return auth(key, email, userName);
  }

  @POST
  public Response createByQueryString(
    @QueryParam("key") String key,
    @QueryParam("email") String email,
    @QueryParam("userName") String userName
  ) {
    logger.info("AuthController createByQueryString(\"{}\", \"{}\", \"{}\")", key, email, userName);
    return auth(key, email, userName);
  }

  private Response auth(String key, String email, String userName) {

    validateApiKey(key);

    SessionLogic sessionLogic = null;

    try {
      sessionLogic = SessionLogic.generateNewSession(key);
    } catch (NoSuchAlgorithmException e) {
      throw new ServerResponseException(uriInfo, null, e);
    }

    try {
      UserLogic userLogic = findUser(email, userName);

      if (userLogic != null) {
        sessionLogic.setUserLogic(userLogic);
      }

      sessionLogic.setClientIp(request.getRemoteAddr());
      sessionLogic.setClientForwarderFor(request.getHeader("X-FORWARDED-FOR"));

      getSessionDAO().save(sessionLogic);
    } catch (TransactionSystemException exc) {
      throw new ServerResponseException(uriInfo, ServerErrorType.INVALID_KEY, exc);
    }

    return Response
      .status(Response.Status.CREATED)
      .location(UriBuilder.fromPath("/" + sessionLogic.getId()).build())
      .entity(getJsonStringFromEntity(sessionLogic.getPojos()))
      .expires(sessionLogic.getExpiration())
      .build();
  }

  private UserLogic findUser(String email, String userName) {
    boolean isEmail = (email != null && email.length() > 0);
    boolean isUserName = (userName != null && userName.length() > 0);

    //optional params
    if (isEmail) {
      UserValidator.validateEmail(email, uriInfo);
    }

    //optional params
    if (isUserName) {
      UserValidator.validateUserName(userName, uriInfo);
    }

    //if both are available
    if (isEmail && isUserName) {
      UserLogic userLogic = userDAO.retrieveByEmailAndName(email, userName, AccountType.PICHEESE);

      if (userLogic == null) {
        throw new ServerResponseException(uriInfo, ServerErrorType.INVALID_AUTH_FOR_USER);
      }

      return userLogic;
    }

    return null;
  }

  @POST
  @Consumes(MediaType.APPLICATION_FORM_URLENCODED)
  @Path("{type}")
  public Response createExternalByForm(
    @PathParam("type") String type,
    @FormParam("key") String key,
    @FormParam("userId") String userId,
    @FormParam("userName") String userName,
    @FormParam("firstName") String firstName,
    @FormParam("lastName") String lastName,
    @FormParam("locale") String locale
  ) {
    logger.info("AuthController createExternalByForm(\"{}\", \"{}\", \"{}\", \"{}\", \"{}\", \"{}\", \"{}\")", type, key, userId, userName, firstName, lastName, locale);
    return authExternal(type, key, userId, userName, firstName, lastName, locale);
  }

  @POST
  @Path("{type}")
  @Consumes(MediaType.APPLICATION_JSON)
  public Response createExternalByJson(
    @PathParam("type") String type,
    JSONObject jsonObject
  ) {

    String key = null;
    String userId = null;
    String userName = null;
    String firstName = null;
    String lastName = null;
    String locale = null;

    try {
      key = jsonObject.getString("key");
    } catch (JSONException e) {
      //do nothing
    }

    try {
      userId = jsonObject.getString("userId");
      userName = jsonObject.getString("userName");
      firstName = jsonObject.getString("firstName");
      lastName = jsonObject.getString("lastName");
      locale = jsonObject.getString("locale");
    } catch (JSONException e) {
      throw new ServerResponseException(uriInfo, ServerErrorType.BAD_REQUEST, e);
    }

    logger.info("AuthController createExternalByJson(\"{}\", \"{}\", \"{}\", \"{}\", \"{}\", \"{}\", \"{}\")", type, key, userId, userName, firstName, lastName, locale);
    return authExternal(type, key, userId, userName, firstName, lastName, locale);
  }

  @POST
  @Path("{type}")
  public Response createExternalByQueryString(
    @PathParam("type") String type,
    @QueryParam("key") String key,
    @QueryParam("userId") String userId,
    @QueryParam("userName") String userName,
    @QueryParam("firstName") String firstName,
    @QueryParam("lastName") String lastName,
    @QueryParam("locale") String locale
  ) {
    logger.info("AuthController createByQueryString(\"{}\", \"{}\", \"{}\", \"{}\", \"{}\", \"{}\", \"{}\")", type, key, userId, userName, firstName, lastName, locale);
    return authExternal(type, key, userId, userName, firstName, lastName, locale);
  }

  private Response authExternal(String type, String key, String userId, String userName, String firstName, String lastName, String locale) {

    validateApiKey(key);
    type = type.toLowerCase();

    if (type.equalsIgnoreCase("fb")) {
//      FacebookAuthValidator fbAuth = new FacebookAuthValidator();
//      if (fbAuth.isValid(userId)) {
//        //TODO
//      }
    }

    SessionLogic sessionLogic = null;

    try {
      sessionLogic = SessionLogic.generateNewSession(key);
      sessionLogic.setClientIp(request.getRemoteAddr());
      sessionLogic.setClientForwarderFor(request.getHeader("X-FORWARDED-FOR"));
    } catch (NoSuchAlgorithmException e) {
      throw new ServerResponseException(uriInfo, null, e);
    }

    try {
      getSessionDAO().save(sessionLogic);
    } catch (TransactionSystemException exc) {
      throw new ServerResponseException(uriInfo, ServerErrorType.INVALID_KEY, exc);
    }

    return Response
      .status(Response.Status.CREATED)
      .location(UriBuilder.fromPath("/" + sessionLogic.getId()).build())
      .entity(getJsonStringFromEntity(sessionLogic.getPojos()))
      .expires(sessionLogic.getExpiration())
      .build();
  }

  @GET
  @Path("/{id}")
  public Response retrieve(@PathParam("id") String id) {
    return super.retrieve(id, uriInfo);
  }

}
