package net.stepniak.api.picheese.controller;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.sun.jersey.core.header.FormDataContentDisposition;
import com.sun.jersey.multipart.FormDataParam;
import net.stepniak.api.entities.SearchResults;
import net.stepniak.api.error.exception.ServerResponseException;
import net.stepniak.api.image.ImageContent;
import net.stepniak.api.image.Resize;
import net.stepniak.api.picheese.logic.PhotoLogic;
import net.stepniak.api.picheese.logic.PhotoRateLogic;
import net.stepniak.api.picheese.logic.UserLogic;
import net.stepniak.api.picheese.repository.PhotoDAO;
import net.stepniak.api.picheese.repository.PhotoRateDAO;
import net.stepniak.api.picheese.services.PicheeseBaseService;
import net.stepniak.api.picheese.validator.PagingValidator;
import net.stepniak.api.picheese.validator.PhotosValidator;
import net.stepniak.api.picheese.validator.UrlValidator;
import net.stepniak.picheese.error.server.ServerErrorType;
import net.stepniak.picheese.error.server.exception.PhotoException;
import net.stepniak.picheese.pojos.v1.CollectionPhoto;
import net.stepniak.picheese.pojos.v1.Photo;
import org.codehaus.jettison.json.JSONException;
import org.codehaus.jettison.json.JSONObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.crypto.codec.Base64;
import org.springframework.stereotype.Service;

import javax.ws.rs.*;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.UriBuilder;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;

@Path("photos")
@Produces(MediaType.APPLICATION_JSON + "; charset=utf-8")
//@Produces(MediaType.APPLICATION_JSON)
@Service
public class PhotoController extends PicheeseBaseService<PhotoLogic> {
  static Logger logger = LoggerFactory.getLogger(PhotoController.class);

  public PhotoController() {
    logger.info("PhotoController constructor()");
  }

  public PhotoDAO getPhotoDAO() {
    return (PhotoDAO) getDao();
  }

  @Autowired
  private PhotoRateDAO photoRateDAO;

  @Autowired
  public void setPhotoDAO(PhotoDAO dao) {
    setDao(dao);
  }

  @GET
  @Path("{id}")
  public Response retrieve(@PathParam("id") Long id) {
    logger.info("PhotoController retrieve(id={})", id);

    validateSessionId("");

    return retrieve(id, uriInfo);
  }

  @GET
  public Response getCollection(
    @QueryParam("sessionId") String sessionId,
    @QueryParam("page") int page,
    @QueryParam("limit") int limit,
    @QueryParam("top") boolean top,
    @QueryParam("lucky") boolean lucky
  ) throws JsonProcessingException {
    logger.info("PhotoController retrieveCollection(lucky={}, top={}, page={}, limit={})", lucky, top, page, limit);

    validateSessionId(sessionId);

    PagingValidator.validatePageNumber(page, uriInfo);
    PagingValidator.validatePageLimit(limit, uriInfo);

    SearchResults<PhotoLogic> entities = null;

    if (top) {
      entities = getPhotoDAO().retrieveTop(page, limit);
    } else if (lucky) {
      entities = getPhotoDAO().retrieveLucky(page, limit);
    } else {
      entities = getPhotoDAO().retrieveByLatest(page, limit);
    }

    CollectionPhoto collectionPhoto = new CollectionPhoto(page, limit, entities.getTotalResults(), entities.getEntities(Photo.class));

    return Response
      .ok()
      .entity(getJsonStringFromEntity(collectionPhoto))
      .build();
  }

  @POST
  @Consumes(MediaType.APPLICATION_FORM_URLENCODED)
  public Response createByForm(
    @FormParam("sessionId") String sessionId,
    @FormParam("title") String title,
    @FormParam("description") String description,
    @FormParam("url") String url
  ) {
    logger.info("PhotoController createByForm(\"{}\", \"{}\")", description, url);

    return upload(sessionId, title, description, url);
  }

  @POST
  @Consumes(MediaType.APPLICATION_JSON)
  public Response createByJson(
    JSONObject jsonObject
  ) {

    String sessionId = null;
    String title = null;
    String description = null;
    String url = null;

    try {
      sessionId = jsonObject.getString("sessionId");
    } catch (JSONException e) {
      //do nothing
    }

    try {
      title = jsonObject.getString("title");
      description = jsonObject.getString("description");
      url = jsonObject.getString("url");
    } catch (JSONException e) {
      throw new ServerResponseException(uriInfo, ServerErrorType.BAD_REQUEST, e);
    }

    logger.info("PhotoController createByJson(\"{}\", \"{}\", \"{}\")", title, description, url);
    return upload(sessionId, title, description, url);
  }

  @POST
  public Response createByQueryString(
    @QueryParam("sessionId") String sessionId,
    @QueryParam("title") String title,
    @QueryParam("description") String description,
    @QueryParam("url") String url
  ) {
    logger.info("PhotoController createByQueryString(\"{}\", \"{}\", \"{}\")", title, description, url);
    return upload(sessionId, title, description, url);
  }

  private Response upload(String sessionId, String title, String description, String url) {
    UserLogic userLogic = getAuthorizedUser(sessionId);

    PhotosValidator.validateDescription(description, uriInfo);
    PhotosValidator.validateTitle(title, uriInfo);
    UrlValidator.validateUrl(url, uriInfo);

//    throw new NotAuthorizedResponseException(uriInfo);  //TODO secure
    PhotoLogic uploadedPhoto = null;

    try {
      byte[] uploadedData = Resize.convertUrlToData(url);
      ImageContent imageContent = new ImageContent(uploadedData);

      uploadedPhoto = new PhotoLogic(userLogic);
      uploadedPhoto.setDescription(description);
      uploadedPhoto.setTitle(title);
      uploadedPhoto.setImageContent(imageContent);
      getPhotoDAO().save(uploadedPhoto);
    } catch (MalformedURLException e) {
      throw new ServerResponseException(uriInfo, ServerErrorType.PARAM_URL, e);
    } catch (PhotoException e) {
      throw new ServerResponseException(uriInfo, ServerErrorType.PARAM_DESCRIPTION, e);
    }

    logger.info("PhotoController upload(\"{}\")", uploadedPhoto.getId());

    return Response
      .status(Response.Status.CREATED)
      .entity(getJsonStringFromEntity(uploadedPhoto.getPojos()))
      .location(UriBuilder.fromPath("/" + uploadedPhoto.getId()).build())
      .build();
  }

  @PUT
  @Consumes(MediaType.APPLICATION_FORM_URLENCODED)
  public Response createWithImgByForm(
    @FormParam("sessionId") String sessionId,
    @FormParam("title") String title,
    @FormParam("description") String description,
    @FormParam("imgBase64") String imgBase64
  ) {
    logger.info("PhotoController createWithImgByForm(\"{}\", \"{}\")", title, description);

    byte[] data = Base64.decode(imgBase64.getBytes());

    return uploadWithImg(sessionId, title, description, data);
  }

  @PUT
  @Consumes(MediaType.MULTIPART_FORM_DATA)
  public Response createWithImgByFormMultipart(
    @FormDataParam("sessionId") String sessionId,
    @FormDataParam("title") String title,
    @FormDataParam("description") String description,
    @FormDataParam("imgFile") FormDataContentDisposition fileDetail,
    @FormDataParam("imgFile") InputStream fileContent
  ) {

    if (fileContent == null) {
      throw new ServerResponseException(uriInfo, ServerErrorType.BAD_REQUEST, new Exception("Missing imgFile param value."));
    }

    ByteArrayOutputStream buffer = new ByteArrayOutputStream();
    int nRead;
    byte[] data = new byte[16384];
    byte[] imageInByte = null;

    try {
      while ((nRead = fileContent.read(data, 0, data.length)) != -1) {
        buffer.write(data, 0, nRead);
      }
      buffer.flush();
      imageInByte = buffer.toByteArray();
      buffer.close();
    } catch (IOException e) {
      throw new ServerResponseException(uriInfo, ServerErrorType.PARAM_IMAGE_BASE64, e);
    }

    logger.info("PhotoController createWithImgByFormMultipart(\"{}\", \"{}\")", title, description);
    return uploadWithImg(sessionId, title, description, imageInByte);
  }

  @PUT
  @Consumes(MediaType.APPLICATION_JSON)
  public Response createWithImgByJson(
    JSONObject jsonObject
  ) {

    byte[] data = null;
    String title = null;
    String description = null;
    String sessionId = null;


    try {
      sessionId = jsonObject.getString("sessionId");
    } catch (JSONException e) {
      //do nothing
    }

    try {
      title = jsonObject.getString("title");
      description = jsonObject.getString("description");
      String imgBase64 = jsonObject.getString("imgBase64");
      data = Base64.decode(imgBase64.getBytes());
    } catch (JSONException e) {
      throw new ServerResponseException(uriInfo, ServerErrorType.BAD_REQUEST, e);
    }

    logger.info("PhotoController createWithImgByJson(\"{}\", \"{}\")", title, description);
    return uploadWithImg(sessionId, title, description, data);
  }

  private Response uploadWithImg(String sessionId, String title, String description, byte[] data) {
    UserLogic userLogic = getAuthorizedUser(sessionId);

    PhotosValidator.validateDescription(description, uriInfo);
    PhotosValidator.validateTitle(title, uriInfo);

    if (data == null || data.length < 1) {
      throw new ServerResponseException(uriInfo, ServerErrorType.PARAM_IMAGE_BASE64);
    }

    PhotoLogic uploadedPhoto = null;

    try {
      ImageContent imageContent = new ImageContent(data);

      uploadedPhoto = new PhotoLogic(userLogic);
      uploadedPhoto.setDescription(description);
      uploadedPhoto.setTitle(title);
      uploadedPhoto.setImageContent(imageContent);

      getPhotoDAO().save(uploadedPhoto);
    } catch (PhotoException e) {
      throw new ServerResponseException(uriInfo, ServerErrorType.PARAM_DESCRIPTION, e);
    }

    logger.info("PhotoController uploadWithImg(\"{}\")", uploadedPhoto.getId());

    return Response
      .status(Response.Status.CREATED)
      .entity(getJsonStringFromEntity(uploadedPhoto.getPojos()))
      .location(UriBuilder.fromPath("/" + uploadedPhoto.getId()).build())
      .build();
  }

  @POST
  @Path("{id}/rate")
  @Consumes(MediaType.APPLICATION_FORM_URLENCODED)
  public Response rateByForm(
    @PathParam("id") Long id,
    @FormParam("sessionId") String sessionId,
    @FormParam("rate") int rate
  ) {

    logger.info("PhotoController rateByForm(\"{}\", \"{}\")", id, rate);
    validateSessionId("");
    return ratePhoto(sessionId, id, rate);
  }

  @POST
  @Path("{id}/rate")
  @Consumes(MediaType.MULTIPART_FORM_DATA)
  public Response rateByFormMultipart(
    @PathParam("id") Long id,
    @FormDataParam("sessionId") String sessionId,
    @FormDataParam("rate") int rate
  ) {

    logger.info("PhotoController rateByFormMultipart(\"{}\", \"{}\")", id, rate);
    validateSessionId(sessionId);
    return ratePhoto(sessionId, id, rate);
  }

  @POST
  @Path("{id}/rate")
  @Consumes(MediaType.APPLICATION_JSON)
  public Response rateByJson(
    @PathParam("id") Long id,
    JSONObject jsonObject
  ) {

    String sessionId = null;
    int rate = 0;

    try {
      sessionId = jsonObject.getString("sessionId");
      rate = jsonObject.getInt("rate");
    } catch (JSONException e) {
      throw new ServerResponseException(uriInfo, ServerErrorType.BAD_REQUEST, e);
    }

    logger.info("PhotoController rateByJson(\"{}\", \"{}\")", id, rate);
    return ratePhoto(sessionId, id, rate);
  }

  private Response ratePhoto(String sessionId, Long id, int rate) {
    PhotosValidator.validateRate(rate, uriInfo);

    UserLogic userLogic = getAuthorizedUser(sessionId);
    PhotoLogic ratedPhoto = get(id);

    PhotoRateLogic photoRateLogic = new PhotoRateLogic(userLogic, ratedPhoto);

    try {
      photoRateDAO.save(photoRateLogic);
    } catch (Exception e) {
      throw new ServerResponseException(uriInfo, ServerErrorType.ALREADY_RATED_PHOTO);
    }

    if (rate > 0) {
      ratedPhoto.increaseScore();
    } else {
      ratedPhoto.decreaseScore();
    }

    return update(ratedPhoto, uriInfo);
  }
}
